﻿using System;
using System.Net.Sockets;

namespace ModBus
{
    class Program
    {
        static void Main(string[] args)
        {
            TcpClient client = new TcpClient();
            client.Connect("192.168.1.69", 502);

            var networkStream = client.GetStream();

            // get the device name at modbus address 0
            ReadRegisters(networkStream, 90, 0, 8);

            byte[] readBuffer = new byte[17 + 8];
            networkStream.Read(readBuffer, 0, readBuffer.Length);

            var deviceName = GetString(readBuffer, 8);
            Console.WriteLine("Device Name: " + deviceName);

            // example of turning on AC Support
            //WriteUInt16(networkStream, 90, 0x0129, 0);
            // example of setting AC Support Voltage to 27V
            //WriteUInt32(networkStream, 90, 0x0150, 27000);

            while (true)
            {
                // get some of the common voltage/power numbers from the SW
                // the modbus memory map can be found on the ComBox website here:
                // https://solar.schneider-electric.com/product/conext-combox/
                float voltage = ReadUInt32(networkStream, 90, 0x004e) * 0.001f;
                Console.WriteLine("Battery voltage: " + voltage + "V");

                float loadPower = ReadUInt32(networkStream, 90, 0x0084);
                Console.WriteLine("Load power: " + loadPower + "W");

                float batteryPower = ReadInt32(networkStream, 90, 0x0052);
                Console.WriteLine("Battery power: " + batteryPower + "W");

                float gridPower = ReadInt32(networkStream, 90, 0x0064);
                Console.WriteLine("Grid power: " + gridPower + "W");

                float totalPower = -(batteryPower + gridPower);
                float wastedPower = totalPower - loadPower;
                Console.WriteLine("Total Power: " + totalPower + "W, wasting: " + wastedPower + "W");

                uint acSupport = ReadUInt16(networkStream, 90, 0x0129);
                Console.WriteLine("AC Support: " + (acSupport == 0 ? "Disabled" : "Enabled"));

                System.Threading.Thread.Sleep(1000);
            }
        }

        private static int ReadInt32(NetworkStream tcpIn, byte modbusAddress, int registerAddress)
        {
            ReadRegisters(tcpIn, modbusAddress, registerAddress, 2);

            byte[] readBuffer = new byte[261];
            tcpIn.Read(readBuffer, 0, 13);

            byte[] sint = new byte[] { readBuffer[10], readBuffer[9], readBuffer[12], readBuffer[11] };
            return BitConverter.ToInt32(sint, 0);
        }

        private static uint ReadUInt32(NetworkStream tcpIn, byte modbusAddress, int registerAddress)
        {
            ReadRegisters(tcpIn, modbusAddress, registerAddress, 2);

            byte[] readBuffer = new byte[261];
            tcpIn.Read(readBuffer, 0, 13);

            return ((uint)readBuffer[11] << 24) | ((uint)readBuffer[12] << 16) | ((uint)readBuffer[9] << 8) | (uint)readBuffer[10];
        }

        private static uint ReadUInt16(NetworkStream tcpIn, byte modbusAddress, int registerAddress)
        {
            ReadRegisters(tcpIn, modbusAddress, registerAddress, 1);

            byte[] readBuffer = new byte[261];
            tcpIn.Read(readBuffer, 0, 11);

            return ((uint)readBuffer[9] << 8) | readBuffer[10];
        }

        private static void WriteUInt16(NetworkStream networkStream, byte modbusAddress, int registerAddress, ushort data)
        {
            byte[] buffer = new byte[12];
            buffer[5] = 6;
            buffer[6] = modbusAddress;
            buffer[7] = 6;

            buffer[8] = (byte)(registerAddress >> 8);
            buffer[9] = (byte)(registerAddress & 0xff);

            buffer[10] = (byte)(data >> 8);
            buffer[11] = (byte)(data & 0xff);

            networkStream.Write(buffer, 0, 12);
            networkStream.Flush();

            byte[] readBuffer = new byte[12];
            networkStream.Read(readBuffer, 0, 12);
        }

        private static void WriteUInt32(NetworkStream networkStream, byte modbusAddress, int registerAddress, ushort data)
        {
            byte[] buffer = new byte[17];
            buffer[5] = 11;
            buffer[6] = modbusAddress;
            buffer[7] = 16;

            buffer[8] = (byte)(registerAddress >> 8);
            buffer[9] = (byte)(registerAddress & 0xff);

            buffer[10] = 0;
            buffer[11] = 2;
            buffer[12] = 4;

            buffer[13] = (byte)((data >> 8) & 0xff);
            buffer[14] = (byte)((data >> 0) & 0xff);
            buffer[15] = (byte)((data >> 24) & 0xff);
            buffer[16] = (byte)((data >> 16) & 0xff);

            networkStream.Write(buffer, 0, 17);
            networkStream.Flush();

            byte[] readBuffer = new byte[12];
            networkStream.Read(readBuffer, 0, 12);
        }

        private static void ReadRegisters(NetworkStream networkStream, byte modbusAddress, int registerAddress, byte length)
        {
            byte[] buffer = new byte[12];
            buffer[5] = 6;
            buffer[6] = modbusAddress;
            buffer[7] = 3;

            buffer[8] = (byte)(registerAddress >> 8);
            buffer[9] = (byte)(registerAddress & 0xff);

            buffer[10] = 0;
            buffer[11] = length;

            networkStream.Write(buffer, 0, 12);
            networkStream.Flush();
        }

        private static string GetString(byte[] bytes, int length)
        {
            // modbus data starts at offset 9
            char[] newString = new char[length];

            for (int i = 0; i < length; i += 2)
            {
                newString[i] = (char)bytes[9 + i + 1];
                newString[i + 1] = (char)bytes[9 + i];
            }

            return new string(newString);
        }
    }
}
